/* pi calculator (c) Terje Mathisen 1997-11-22 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <windows.h>
#include <mmsystem.h>           // Used for millisecond timing!

typedef unsigned long ul;

ul blocks = 100;                // Default # of blocks in a number

ul divmask = (ul) -1;

void ulprint(char *head, ul *nr, long decimals);

#pragma warning (disable : 4035)

ul ldiv_asm(ul *dst, ul *src, ul len, ul divisor, ul carry)
{
//	ul res;
	__asm {
		mov esi,[src]
		mov edi,[dst]
		mov ecx,[len]
		sub esi,4
		mov ebx,[divisor]
		mov edx,[carry]
		 jmp first
next:
		div ebx
		mov [edi+ecx*4],eax
first:
		mov eax,[esi+ecx*4]
		dec ecx
		 jnz next

		div ebx
		mov [edi],eax
//		mov [res],edx
		mov eax,edx
	}
//	return res;
}

ul lmul_asm(ul *dst, ul *src, ul len, ul multiplier, ul carry)
{
//	ul res;
	__asm {
		mov esi,[src]
		mov edi,[dst]
		mov ecx,[len]
		sub edi,esi
		mov ebx,[carry]
		push ebp
		sub edi,4
		mov ebp,[multiplier]
next:
		mov eax,[esi]
		lea esi,[esi+4]
		mul ebp
		add eax,ebx
		mov ebx,0
		adc ebx,edx
		mov [edi+esi],eax
		dec ecx
		 jnz next
			
		pop ebp
//		mov [res],ebx
		mov eax,ebx
	}
//	return res;
}

ul ladd_asm(ul *dst, ul *src, ul len)
{
//	ul res;
	__asm {
		mov esi,[src]
		mov edi,[dst]
		mov ecx,[len]
		sub esi,edi
next:
		mov eax,[edi]
		mov edx,[edi+esi]
		adc eax,edx
		mov [edi],eax
		lea edi,[edi+4]
		dec ecx
		 jnz next

		sbb eax,eax
		neg eax
//		mov [res],eax
	}
//	return res;
}

ul lsub_asm(ul *dst, ul *src, ul len)
{
//	ul res;
	__asm {
		mov esi,[src]
		mov edi,[dst]
		mov ecx,[len]
		sub esi,edi
next:
		mov eax,[edi]
		mov edx,[edi+esi]
		sbb eax,edx
		mov [edi],eax
		lea edi,[edi+4]
		dec ecx
		 jnz next

		sbb eax,eax
//		neg eax
//		mov [res],eax
	}
//	return res;
}
#pragma warning (default : 4035)


#ifndef NDEBUG
ul divs = 0, muls = 0, adds = 0, subs = 0, maxdiv = 0;
#endif

ul ldiv(ul *dst, ul *src, ul divisor)
{
#ifndef NDEBUG
	divs++;
#endif

	ul len = src[0], res = 0;
	
	if (len) {
		res = ldiv_asm(dst+1, src+1, len, divisor, 0);
		if (dst[len] == 0) len--;
	}
	dst[0] = len;
	
	return res;
}

ul lmul(ul *dst, ul *src, ul multiplier, ul carry)
{
#ifndef NDEBUG
	muls++;
#endif

	ul len = src[0], res = 0;
	
	if (len) {
		res = lmul_asm(dst+1, src+1, len, multiplier, carry);
		
		if (res) {
			len++;
			dst[len] = res;
		}
	}
	dst[0] = len;
	
	return res;
}

ul ladd(ul *dst, ul *src)
{
#ifndef NDEBUG
	adds++;
#endif

	ul len = src[0], res = 0, sum;
	
	if (len) {
		res = ladd_asm(dst+1, src+1, len);

		if (res) {
			res = 0;
			do {
				if (len >= blocks) {
					res = 1;
					break;
				}
				len++;
			
				sum = dst[len] + 1;
				dst[len] = sum;
			} while (sum == 0);
		}
		
		if (len > dst[0]) 
			dst[0] = len;
	}
	return res;
}

ul lsub(ul *dst, ul *src)
{
#ifndef NDEBUG
	subs++;
#endif

	ul len = src[0], dlen = dst[0], res, diff;
	
	if (len) {
		res = lsub_asm(dst+1, src+1, len);

		if (res) {
			res = 0;
			do {
				if (len >= blocks) {
					res = 1;
					break;
				}
				len++;
			
				diff = dst[len] - 1;
				dst[len] = diff;
			} while (diff == (ul) -1);
		}
		if (len > dst[0]) 
			dst[0] = len;
	}
	return 0;
}

ul *load(ul *dst, ul nr)
{
	memset(dst, 0, sizeof(ul)*(blocks+1));
	dst[blocks] = nr; dst[0] = blocks;
	
	return dst;
}

ul *init(void)
{
	ul *t;
	
	t = (ul *) malloc(sizeof(ul) * (blocks+1));
	load(t, 0);
	
	return t;
}

ul *copy(ul *dst, ul *src)
{
	memcpy(dst, src, sizeof(ul) * (blocks+1));
	
	return dst;
}

ul *acotan(ul n, ul *dst)
{
	ul n2 = n * n, divisor = 3, maxdivisor, r;
    ul *t1, *t2, n4;
	
	n4 = n2*n2;
	maxdivisor = 0xffffffff / n2 - 1;

	maxdivisor &= divmask;
	
    t1 = load(init(), 1); t2 = init();
	
	ldiv(t1, t1, n);
	copy(dst, t1);
//	ulprint("1/num = ", dst, 50);
	
	do {
		if (divisor < maxdivisor) {
			r =ldiv(t2, t1, n2*divisor);
			lmul(t1, t2, divisor, r);
			divisor += 2;
			lsub(dst, t2);
			
			r =ldiv(t2, t1, n2*divisor);
			lmul(t1, t2, divisor, r);
			divisor += 2;
			ladd(dst, t2);
		}
		else {
			ldiv(t1, t1, n2);
//	ulprint("x/n2 = ", t1, 50);
			ldiv(t2, t1, divisor);
//	ulprint("x/n2/d=", t2, 50);
			divisor += 2;
			lsub(dst, t2);
//	ulprint("diff = ", dst, 50);
			
			ldiv(t1, t1, n2);
//	ulprint("x/n2 = ", t1, 50);
			ldiv(t2, t1, divisor);
//	ulprint("x/n2/d=", t2, 50);
			divisor += 2;
			ladd(dst, t2);
//	ulprint("sum  = ", dst, 50);
		}
	} while (t2[0]);
	
	free(t1); free(t2);

#ifndef NDEBUG
	if (divisor > maxdiv) maxdiv = divisor;
#endif
	
	return dst;
}

ul *calc_pi_5_239(ul *dst)
{
	ul *t;
	
	t = init();
	
	acotan(5, dst);
	
	ladd(dst, dst);
	ladd(dst, dst);
//	ulprint("4x1/5= ", dst, 50);
	
	acotan(239, t);
	
	lsub(dst, t);
//	ulprint("pi/4 = ", dst, 50);

	ladd(dst, dst);
	ladd(dst, dst);
	
	free(t);
	
	return dst;
}

ul *calc_pi(ul *dst)
{
/*	Formel hentet fra:
	http://www.ccsf.caltech.edu/~roy/pi.formulas.html

	0.856797 1 :: 44:57 7:239 -12:682 24:12943 
*/
	ul *t;
	
	t = init();
	
	acotan(57, dst);
	lmul(dst, dst, 4*44, 0);

	acotan(239, t);
	lmul(t, t, 7*4, 0);
	ladd(dst, t);

	acotan(682, t);
	lmul(t, t, 12*4, 0);
	lsub(dst, t);

	acotan(12943, t);
	lmul(t, t, 24*4, 0);
	ladd(dst, t);

	free(t);
	
	return dst;
}

#define BLOCKSIZE	5
#define LINESIZE	50

void ulprint(char *head, ul *nr, long decimals)
{
	ul *t, len = nr[0];
	
	t = copy(init(), nr);
	while (len < blocks) {
		t[len+1] = 0;
		len++;
	}
	printf("%s%d.\n", head, t[len]);
	
	int block = BLOCKSIZE, pos = LINESIZE;
	char line[LINESIZE*2], *pline = line;

	do {
		char buf[10];

		t[0] = len-1;
		lmul(t, t, 1000000000, 0);
		sprintf(buf,"%09d", t[len]);

		for (int i = 0; i < 9; i++) {
			*pline++ = buf[i];
			if (--decimals == 0) break;

			if (--pos == 0) {
				*pline = '\0'; puts(line);
				pline = line; pos = LINESIZE; block = BLOCKSIZE;
			}
			else if (--block == 0) {
				*pline++ = ' ';
				block = BLOCKSIZE;
			}
		}

	} while (decimals);
	*pline = '\0'; puts(line);
	
	free(t);
}

int main(int argc, char *argv[])
{
	int digits = 25, dispdigits;
	
	if (argc > 1) {
		digits = atoi(argv[1]);
	}
	dispdigits = digits;
	if (argc > 2) {
		dispdigits = atoi(argv[2]);
	}
	
	blocks = (int) (digits / 9.632959861247397) + 5;
	
	ul *pi = init();
	
	timeBeginPeriod(1);
	ul t = timeGetTime();
	calc_pi(pi);
	timeEndPeriod(1);
	t = timeGetTime() - t;
	
	ulprint("  pi = ", pi, dispdigits);

	fprintf(stderr, "Time to calculate %d digits = %1.3f seconds\n", 
		digits, t * 0.001);

#ifndef NDEBUG
	ul *pi_5 = init();

	divmask = 0;
	
	timeBeginPeriod(1);
	t = timeGetTime();
	calc_pi(pi);
	timeEndPeriod(1);
	t = timeGetTime() - t;
	
	fprintf(stderr, "TDiv to calculate %d digits = %1.3f seconds\n", 
		digits, t * 0.001);

	timeBeginPeriod(1);
	t = timeGetTime();
	calc_pi_5_239(pi_5);
	timeEndPeriod(1);
	t = timeGetTime() - t;

	fprintf(stderr, "5/239   calculate %d digits = %1.3f seconds\n", 
		digits, t * 0.001);

	if (memcmp(&pi[2], &pi_5[2], sizeof(ul)*(blocks-1))) {
		fprintf(stderr,"Not equal results!\n");
		ulprint("pi_5 = ", pi, dispdigits);
		for (int i = blocks; (i >= 0) && (pi[i] != pi_5[i]); i--) ;
		fprintf(stderr, "%d different blocks!\n", blocks - i);
	}
	else 
		fprintf(stderr, "All OK!\n");
	
	free(pi_5);

	fprintf(stderr,"divs = %u, muls = %u, adds = %u, subs = %u, maxdiv = %u\n",
		divs, muls, adds, subs, maxdiv);

#endif
	free(pi); 

	return 0;
}
